from __future__ import annotations

import typing

import requests
from bs4 import BeautifulSoup


class MonsnodeParser:    # pylint: disable=too-few-public-methods
    TARGET_URL = 'https://monsnode.com/'
    timeout: float = 5000

    def is_target_reachable(self) -> bool:
        try:
            requests.get(self.TARGET_URL, timeout=self.timeout)
            return True
        except requests.exceptions.ConnectionError:
            return False

    def get_video(self, link: str) -> tuple[str, str]:
        request = requests.get(link, timeout=self.timeout)
        soup = BeautifulSoup(request.text, 'html.parser')

        video_name = self._parse_video_name(soup)
        link = self._parse_video_url(soup)

        return video_name, link

    def _parse_video_name(self, soup: BeautifulSoup) -> str:

        uploader_name = soup.select('body > div:nth-child(2) > div:nth-child(2) > div:nth-child(5) > b')[0].text
        title = soup.select('body > div:nth-child(2) > div:nth-child(2) > div:nth-child(2)')[0].text
        title = title.replace('/', '_')
        title = title.replace('\n', '')
        return f'{uploader_name} - {title}.mp4'

    def _parse_video_url(self, soup: BeautifulSoup) -> str:
        video_path = typing.cast(str,
                                 soup.select('body > div:nth-child(2) > div:nth-child(1) > a:nth-child(2)')[0]['href'])
        return self.TARGET_URL + video_path



from typing import Optional

import youtube_dl

from .execute_parallel import execute_parallel


class YoutubeDLWrapper:    # pylint: disable=too-few-public-methods
    """A YoutubeDL wrapper class for supporting python embedded multi-processing"""

    def _youtube_dl_download_video(self, link: str, youtube_dl_option: Optional[dict[str, str]] = None) -> None:
        with youtube_dl.YoutubeDL(youtube_dl_option) as youtube_dl_downloader:
            youtube_dl_downloader.download([link])

    def download(self, links: list[str], youtube_dl_option: Optional[dict[str, str]] = None) -> None:
        arguments = [(link, youtube_dl_option) for link in links]
        execute_parallel(self._youtube_dl_download_video, arguments)


youtube_dl_wrapper = YoutubeDLWrapper()    # singleton



import time
from typing import Optional

from playwright.sync_api import Error, Page, Request


class TwitterCrawler:
    page: Page

    def __init__(self, page: Page):
        self.page = page

    @property
    def page_current_height(self) -> int:
        current_height: int = self.page.evaluate('(window.innerHeight + window.scrollY)')
        return current_height

    def login(self, username: str, password: str, timeout: Optional[float] = 10000) -> None:
        print('Logging in...')
        self.page.goto('https://twitter.com/i/flow/login')
        self.page.wait_for_selector('label:has-text(\"Phone, email, or username\") div >> nth=1')
        self.page.get_by_label('Phone, email, or username').click()
        self.page.get_by_label('Phone, email, or username').fill(username)
        self.page.get_by_label('Phone, email, or username').press('Enter')
        try:
            self.page.get_by_label('Password').fill(password, timeout=timeout)
        except Error as error:
            raise ValueError('Login failed; Maybe username or password is incorrect?') from error
        self.page.get_by_label('Password').press('Enter')
        self.page.wait_for_url('https://twitter.com/home')

    def get_all_liked_tweets(self, username: str, scroll_timeout: float = 0.8) -> list[str]:
        """Get the username's all liked tweets
        Returns the list of links of liked tweets
        """
        return self.get_liked_tweets_until(
            username, 'nothing', scroll_timeout
        )    # 'nothing' was intended because the given `until_link` would be never found on the links list

    def get_liked_tweets_until(self, username: str, until_link: str, scroll_timeout: float = 0.8) -> list[str]:
        """Scrolling down the list of liked tweets until the given `until_link` found
        Returns the list of links of liked tweets
        """
        self._goto_liked_tweets(username)
        links: list[str] = []

        previous_height = self.page_current_height
        while True:
            # 1. scroll down
            # 2. get the link of tweets in the current screen(tweets are not reachable if it's not rendering)
            # 3. break if page reaches to the bottom or the given `until_link` found

            self.page.mouse.wheel(0, 1500)
            time.sleep(scroll_timeout)    # wait for mouse wheel to scroll down
            is_page_bottom = self.page_current_height == previous_height
            if is_page_bottom:
                break
            previous_height = self.page_current_height

            new_links = self._get_article_links_in_current_screen()
            links.extend(new_links)
            links = list(set(links))

            print(f'Found {len(links)} liked tweets.')

            if until_link in links:
                break

        return links

    def get_recent_liked_tweet(self, username: str) -> str:
        self._goto_liked_tweets(username)
        return self._get_article_links_in_current_screen()[0]

    def get_video_of_tweet(self, link: str, timeout: Optional[float] = 5000) -> list[tuple[str, str]]:
        video_links: list[str] = []

        def _request_m3u8_capture_handler(request: Request) -> None:
            if 'm3u8' in request.url:
                video_links.append(request.url)

        self.page.on('request', _request_m3u8_capture_handler)
        self.page.goto(link)
        try:
            self.page.wait_for_selector('video', timeout=timeout)
        except Error:
            return []

        return [(f'{self._parse_tweet_name()}_{index}.mp4', link) for index, link in enumerate(video_links)]

    def _parse_tweet_name(self) -> str:
        uploader = self.page.get_by_test_id('primaryColumn').get_by_role('link').nth(0).inner_text().strip()
        content = self.page.get_by_role('article').get_by_test_id('tweetText').nth(0).inner_text().strip()
        return f'{uploader} - {content}'

    def _goto_liked_tweets(self, username: str) -> None:
        self.page.goto(f'https://twitter.com/{username}/likes')
        self.page.wait_for_selector('article')

    def _get_article_links_in_current_screen(self) -> list[str]:
        links: list[str] = []

        while True:
            articles = self.page.locator('article')
            article_length = articles.count()
            try:
                links = [
                    'https://twitter.com' +
                    (articles.nth(i).locator('div').locator('a').nth(3).get_attribute('href', timeout=500) or '')
                    for i in range(article_length)
                ]
                break
            except Error:    # if articles in the page are not reachable
                self.page.mouse.wheel(0, 500)    #  scrolling down to refresh the articles

        return links



# pylint: disable=useless-import-alias

from .monsnode_parser import MonsnodeParser as MonsnodeParser
from .platform_video_downloader import \
    PlatformVideoDownloader as PlatformVideoDownloader
from .twitter_crawler import TwitterCrawler as TwitterCrawler
from .video_downloader import VideoDownloader as VideoDownloader



# pylint: disable=redefined-outer-name
import pytest
import requests

from ..monsnode_parser import MonsnodeParser


def test_is_target_reachable(monsnode_parser: MonsnodeParser) -> None:
    # given
    expected_target_reachable = False
    try:
        requests.get(monsnode_parser.TARGET_URL, timeout=monsnode_parser.timeout)
        expected_target_reachable = True
    except requests.exceptions.ConnectionError:
        expected_target_reachable = False

    # when
    actual_target_reachable = monsnode_parser.is_target_reachable()

    # then
    assert actual_target_reachable == expected_target_reachable


@pytest.mark.skipif(not MonsnodeParser().is_target_reachable(), reason='Monsnode is currently not reachable')
def test_get_video_name(monsnode_parser: MonsnodeParser) -> None:
    # given
    target_video_link = 'https://monsnode.com/v1506575871309589251'

    # when
    video_name, video_link = monsnode_parser.get_video(target_video_link)

    # then
    assert video_name == 'ウォーター(@waterpokepwpr) - 剣盾ずっとやってきたけど、こんな経験初めて。 https:__t.co_LT6dHwYitY.mp4'
    assert video_link == 'https://monsnode.com/redirect.php?v=13768280'


@pytest.fixture
def monsnode_parser() -> MonsnodeParser:
    return MonsnodeParser()



import multiprocessing
import multiprocessing.dummy
from typing import Any, Callable


def execute_parallel(func: Callable[[str], None], *args: Any, **kwargs: Any) -> None:
    pool = multiprocessing.dummy.Pool(multiprocessing.cpu_count())
    pool.starmap(func, *args, **kwargs)



from typing import Optional

from playwright.sync_api import sync_playwright
from youtube_dl.utils import DownloadError, YoutubeDLError

from .monsnode_parser import MonsnodeParser
from .twitter_crawler import TwitterCrawler
from .utils.youtube_dl_wrapper import youtube_dl_wrapper


class PlatformVideoDownloader:
    monsnode_parser: MonsnodeParser
    video_output_path: str = 'videos'

    def __init__(self, video_output_path: str = 'videos', monsnode_parser: MonsnodeParser = MonsnodeParser()):
        self.video_output_path = video_output_path
        self.monsnode_parser = monsnode_parser

    def download_monsnode_video(self, link: str) -> None:
        video_filename, video_link = self.monsnode_parser.get_video(link)
        youtube_dl_option = self._make_youtube_dl_option(video_filename)
        youtube_dl_wrapper.download([video_link], youtube_dl_option)

    def download_twitter_video(
        self,
        link: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        try:
            youtube_dl_option = self._make_youtube_dl_option()
            youtube_dl_wrapper.download([link], youtube_dl_option)
        except YoutubeDLError as exception:
            if not username or not password:
                raise Exception('Username and password are required to download private Twitter video') from exception

            videos_info = self._get_twitter_private_videos_info(link, username, password)
            for video_filename, video_link in videos_info:
                youtube_dl_option = self._make_youtube_dl_option(video_filename)
                youtube_dl_wrapper.download([video_link], youtube_dl_option)

    def _make_youtube_dl_option(self, video_filename: Optional[str] = None) -> dict[str, str]:
        youtube_dl_output_template = \
            f'{self.video_output_path}/{video_filename}' \
                if video_filename else \
            f'{self.video_output_path}/%(title)s.%(ext)s'
        return {'format': 'bestvideo/best', 'outtmpl': youtube_dl_output_template}

    def _get_twitter_private_videos_info(
        self,
        target_link: str,
        username: str,
        password: str,
    ) -> list[tuple[str, str]]:
        with sync_playwright() as playwright_sync:
            browser = playwright_sync.webkit.launch(headless=False)
            page = browser.new_page()
            crawler = TwitterCrawler(page)
            crawler.login(username, password)
            videos_info = crawler.get_video_of_tweet(target_link)
            page.close()
            browser.close()
        return videos_info



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['twitter_video_tools',
 'twitter_video_tools.tests',
 'twitter_video_tools.utils']

package_data = \
{'': ['*']}

install_requires = \
['bs4>=0.0.1,<0.0.2',
 'playwright>=1.27.1,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'youtube_dl>=2021.12.17,<2022.0.0']

setup_kwargs = {
    'name': 'twitter-video-tools',
    'version': '2022.11.17.post3',
    'description': 'Twitter Video Tools is a multi-processing supported video downloader, supports videos from twitter (or specific user from twitter) && monsnode.',
    'long_description': '# Twitter Video Tools\n[![PyPI version](https://badge.fury.io/py/twitter-video-tools.svg)](https://badge.fury.io/py/twitter-video-tools)\n[![Test](https://github.com/code-yeongyu/twitter_video_tools/actions/workflows/test.yaml/badge.svg?branch=master)](https://github.com/code-yeongyu/twitter_video_tools/actions/workflows/test.yaml)\n[![codecov](https://codecov.io/gh/code-yeongyu/twitter_video_tools/branch/master/graph/badge.svg?token=97K8BBWOH7)](https://codecov.io/gh/code-yeongyu/twitter_video_tools)\n\n- A multi-processing supported video downloader\n- supports downloading videos from twitter (or specific user from twitter) && monsnode.\n\n## Install\n\n### with PIP\n\n```sh\npip install twitter-video-tools\n```\n\n### with Poetry (Recommended)\n\n```sh\npoetry add twitter-video-tools\n```\n\n## Contribution\n\n### Prerequisites\n\n- Python 3.9\n- poetry\n- code editor (vscode recommended)\n\n### Quick Info of setups\n\n- strict type checking using mypy\n- amazing linters & formatters (`yapf`, `pylint`, `isort`)\n  - `unify` for forcing single-quote\n- unit test using `pytest`\n- vscode launch & formatting setups\n\n### All-in-one\n\n```sh\ngh repo clone code-yeongyu/twitter_video_tools\npython3 -m pip install poetry\npoetry install # install dependencies\ncode --install-extension emeraldwalk.RunOnSave # to force single quote\n```\n\nDone!\n\n### Test\n\n```sh\npoetry shell\ninv test\n```\n',
    'author': 'None',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/code-yeongyu/twitter_video_tools',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import Optional

from twitter_video_tools.platform_video_downloader import \
    PlatformVideoDownloader
from twitter_video_tools.utils import execute_parallel

from .monsnode_parser import MonsnodeParser


class VideoDownloader:
    """YoutubeDL wrapper with extra features"""
    username: Optional[str] = None
    password: Optional[str] = None
    platform_video_downloader: PlatformVideoDownloader

    def __init__(
            self,
            username: Optional[str] = None,
            password: Optional[str] = None,
            platform_video_downloader: PlatformVideoDownloader = PlatformVideoDownloader('videos', MonsnodeParser()),
    ):
        self.username = username
        self.password = password
        self.platform_video_downloader = platform_video_downloader

    def download_videos(self, links: list[str]) -> None:
        arguments = [(link, ) for link in links]
        execute_parallel(self.download_video, arguments)

    def download_video(self, link: str) -> None:
        if 'monsnode.com' in link:
            self.platform_video_downloader.download_monsnode_video(link)
            return
        self.platform_video_downloader.download_twitter_video(link, self.username, self.password)
